#define GLOBAL_DATA 1
/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : aggregator.c                                             */
/* Output name     : aggregator.dll                                           */
/*                   (NOTE: must be renamed to aggregator.lil and moved to    */
/*                   <install dir>\bin for use with MQSI V2.0 broker          */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\include\plugin        */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : This pair of plug-in nodes create and manage an          */
/*                   aggregate message group.                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Apr 2000                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <aggregator.h> /* Specific header for this dll                            */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <plugin.h>   /* common plugin node functions */
#include <BipCni.h>	  /* Function prototypes for Plug-in Node functions     */
#include <log.h>


/******************************************************************************/
/* Node Implementation Functions                                              */
/*                                                                            */
/* These are mandatory and MUST be implemented by the programmer.             */
/*                                                                            */
/* createNodeContext                                                          */
/* deleteNodeContext                                                          */
/* evaluate                                                                   */
/* getAttribute                                                               */
/* getAttributeName                                                           */
/* setAttribute                                                               */
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* createNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  factoryObject (input) - address of factory object that owns the node      */
/*                          being created.                                    */
/*  nodeName      (input) - name of node being created.                       */
/*  nodeObject    (input) - address of the node object that has just been     */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to the local context.                                             */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates any plug-in context for an instance of the node object. This is    */
/* invoked by the broker whenever a node instance is constructed. Nodes are   */
/* constructed whenever a messageflow containing the node is deployed by the  */
/* broker.                                                                    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Optionally verify the name of the node is supported by the factory.     */
/* 2. Allocate node instance specific data areas. e.g. For attribute data or  */
/*    terminals.                                                              */
/* 3. Perform any additional resource acquisition or initialisation that is   */
/*    required for the processing of the node.                                */
/* 4. Return the address of the context to the calling function.              */
/*                                                                            */
/******************************************************************************/
CciContext *createNodeContext(CciFactory *factoryObject, 
                              CciChar *nodeName, 
                              CciNode *nodeObject)
{
    int rc;
	static char *functionName="createNodeContext()";
    context_t *contextP = NULL;
	CciChar messageFlowNameUnicode[MAX_STRING_SIZE + 1];
	char messageFlowNameMB[MAX_STRING_SIZE + 1];
	messageFlow_t *mfP;
    CNI_BROKER_INFO_ST brokerInfo;
	messageFlow_t defaultMsgFlow = {MESSAGE_FLOW_DEFAULT},
		          *messageFlowAnchorP;

    TRACE_ENTRY("createNodeContext", ("factoryObject=%.8x, nodeName=%S, "
		                              "nodeObject=%.8x",
									  factoryObject,
									  nodeName,
									  nodeObject));

	/* initialise the broker info structure */
	ALLOCATE_STRING_STORAGE(brokerInfo.brokerName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.executionGroupName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.messageFlowName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.queueManagerName, 300);

	/* get the name of our queue manager */
    TRACE_MID(TRACE_INFO, ("About to call cniGetBrokerInfo"));

	cniGetBrokerInfo(0,           /* retcode does not appear to be set */
			         nodeObject,
					 &brokerInfo);

	sprintf(messageFlowNameMB, "%S", brokerInfo.messageFlowName.buffer);
	wcscpy(messageFlowNameUnicode, brokerInfo.messageFlowName.buffer);
    TRACE_MID(TRACE_INFO, ("Extracted messageFlowName is '%s'",
			               messageFlowNameMB));

	/* Is this the first aggregate for this Message Flow */
	threadMutexLock(aggregateChainMutex);
	mfP = findMessageFlow(messageFlowNameUnicode);

	if (!mfP)
	{
		/* create new message flow entry */
		mfP = (messageFlow_t *)malloc(sizeof(messageFlow_t));
		memcpy(mfP, &defaultMsgFlow, sizeof(messageFlow_t));
		mfP->name = UCDup(messageFlowNameUnicode);

		/* add to message flow chain */
		messageFlowAnchorP = &messageFlowAnchor;
		INSERT_AFTER_IN_LINKED_LIST(messageFlowAnchorP, mfP);
	}
	threadMutexUnlock(aggregateChainMutex);

    TRACE_MID(TRACE_INFO, ("Current messageFlow entry at %.8x", mfP));

	if (!brokerInfoObtained)
	{
		/* copy the process specific information, both as single and Unicode strings */
		sprintf(brokerName, "%S", brokerInfo.brokerName.buffer);
		wcscpy(brokerNameUC, brokerInfo.brokerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted brokerName is '%s'",
			                   brokerName));

		sprintf(executionGroupName, "%S", brokerInfo.executionGroupName.buffer);
		wcscpy(executionGroupNameUC, brokerInfo.executionGroupName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted executionGroupName is '%s'",
			                   executionGroupName));

		sprintf(queueManagerName, "%S", brokerInfo.queueManagerName.buffer);
		wcscpy(queueManagerNameUC, brokerInfo.queueManagerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted queueManagerName is '%s'",
			                   queueManagerName));

		brokerInfoObtained = TRUE;
	}

	FREE_STRING_STORAGE(brokerInfo.brokerName);
	FREE_STRING_STORAGE(brokerInfo.executionGroupName);
	FREE_STRING_STORAGE(brokerInfo.messageFlowName);
	FREE_STRING_STORAGE(brokerInfo.queueManagerName);

    /* Allocate a pointer to the local context */
    contextP = malloc(sizeof(context_t));

    /* If malloc was successful */
    if (!contextP)
	{
        /* Error handling if malloc failed */
        TRACE_RETURN_TYPE(RSN_002, void *, contextP);
	}

    TRACE_MID(TRACE_INFO, ("Context structure allocated at %.8x", contextP));

    /* initialise returned storage for context anchor */
    memset (contextP, 0, sizeof(context_t)); /* Clear the context area */
    contextP->nodeObject = nodeObject;
	contextP->messageFlowName = mfP->name;

    /* Verify the name of the node is supported by the factory */
    if (!wcscmp(nodeName, AGGREGATE_CREATE_NODE))
	{
        TRACE_MID(TRACE_INFO, ("<==================================>"));
        TRACE_MID(TRACE_INFO, ("%s node current", AGGREGATE_CREATE_NODE_EXTERNAL_NAME_MB));
        TRACE_MID(TRACE_INFO, ("<==================================>"));

		/* aggregate creation node */
		contextP->nodeType = AGGREGATE_CREATE;
        contextP->nodeName = UCDup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_AGGREGATE_NAME, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_SEQUENCE_NUMBER, CCI_VALUE_TYPE_INTEGER);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"label", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"userTraceLevel", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"traceLevel", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"userTraceFilter", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"traceFilter", CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has a single input and 2 output terminals */
		insertInputTerminalListEntry(contextP, L"in");

		insertOutputTerminalListEntry(contextP, L"out");
		insertOutputTerminalListEntry(contextP, L"failure");

	}
	else
    if (!wcscmp(nodeName, AGGREGATE_REPLY_NODE))
	{
        TRACE_MID(TRACE_INFO, ("<=================================>"));
        TRACE_MID(TRACE_INFO, ("%s node current", AGGREGATE_REPLY_NODE_EXTERNAL_NAME_MB));
        TRACE_MID(TRACE_INFO, ("<=================================>"));

		/* aggregate reply node */
		contextP->nodeType = AGGREGATE_REPLY;
        contextP->nodeName = UCDup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_AGGREGATE_NAME, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_REPLY_TIMEOUT_SECONDS, CCI_VALUE_TYPE_INTEGER);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_REPLY_TO_QUEUE, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_SEPARATOR_STRING, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_HEX_SEPARATOR_STRING, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_AGGREGATE_REPLY_DESTINATION, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      ATTR_PERSISTENT, CCI_VALUE_TYPE_BOOL);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"label", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"userTraceLevel", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"traceLevel", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"userTraceFilter", CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      L"traceFilter", CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has two input terminals and 3 output terminals */
		insertInputTerminalListEntry(contextP, L"in");
		insertInputTerminalListEntry(contextP, L"timerIn");

		insertOutputTerminalListEntry(contextP, L"out");
		insertOutputTerminalListEntry(contextP, L"timedOut");
		insertOutputTerminalListEntry(contextP, L"lateReplies");
		insertOutputTerminalListEntry(contextP, L"failure");
	}
	else
	{
	    /* Error handling if name of node is not supported by factory */
	    free(contextP);
	}

	/* 
	 * If this is the first aggregator node encountered, create the
	 * alarm thread  
	 */
	if (!alarmThreadCreated)
	{
		rc = createAlarmThread(contextP);

		if (rc != TT_SUCCESS)
		{
			TRACE_RETURN_TYPE(RSN_003, void *, contextP);
		}
	}

    /* Return the address of the context to the calling function */
    TRACE_RETURN_TYPE(RSN_004, void *, contextP);
}

/******************************************************************************/
/* createAlarmThread                                                          */
/******************************************************************************/
int createAlarmThread(IN context_t *contextP)
{
	char errorMessage[1000];
	static char *functionName="createAlarmThread()";
	int rc;

    TRACE_ENTRY("createAlarmThread", ("contextP=%.8x",
		                              contextP));

    /* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock alarmThreadSyncMutex"));
	threadMutexLock(alarmThreadSyncMutex);

	/* check again in case we lost the race... */
	if (alarmThreadCreated)
	{
	 	TRACE_MID(TRACE_INFO, ("Lost the race to create alarm thread"));
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_005, TT_SUCCESS);
	}
		
	/* create the alarm thread */
	TRACE_MID(TRACE_INFO, ("About to create alarm thread"));
	if (threadCreateThread((threadFunction_t)processAlarms,
						  (threadParameter_t)contextP,
						   &alarmThreadHandle) != THREAD_OK)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_006, TT_FAILED);
	}
	TRACE_MID(TRACE_INFO, ("Alarm thread (%d) successfully created",
						   threadGetThreadId(alarmThreadHandle)));

	/* wait up to 10 seconds for child thread to signal us that he's up and running */
	TRACE_MID(TRACE_INFO, ("About to wait for alarm thread to initialise"));
	
	rc = threadCondTimedWait(alarmThreadSyncCondVar, 10000);

	if (rc != THREAD_OK && rc != THREAD_TIMED_OUT)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_007, TT_FAILED);
	}

	/* check response from alarm thread */
	if (!alarmThreadInitialised)
	{
		sprintf(errorMessage, "Alarm thread failed to initialise.  Reason='%s'",
							  alarmThreadErrorMessage);
		TRACE_MID(TRACE_ERROR, ("%s", errorMessage));

		/* Output a log message */
		cciLog(&rc,                       /* returnCode */
			   CCI_LOG_ERROR,       /* type */
			   __FILE__,                  /* file */
			   __LINE__,                  /* line */
			   functionName,              /* function */
			   PLUGIN_MESSAGE_CATALOGUE,    /* messageSource */
			   MJB_ALARM_THREAD_INIT_FAILED, /* messageNumber */ 
			   errorMessage,              /* traceText */
#if defined(_AIX)
			   brokerName,
#endif
			   errorMessage,
			   (char *)0);
	}
	else
	{
		alarmThreadCreated = TRUE;

		TRACE_MID(TRACE_INFO, ("Alarm thread initialised ok"));
	}

	TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
	threadMutexUnlock(alarmThreadSyncMutex);
	TRACE_RETURN(RSN_008, TT_SUCCESS);
}

/******************************************************************************/
/* createLogThread                                                            */
/******************************************************************************/
int createLogThread(IN context_t *contextP)
{
	char errorMessage[1000];
	static char *functionName="createLogThread()";
	int rc;

    TRACE_ENTRY("createLogThread", ("contextP=%.8x",
		                             contextP));

    /* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock logThreadSyncMutex"));
	threadMutexLock(logThreadSyncMutex);

	/* check again in case we lost the race... */
	if (logThreadCreated)
	{
		TRACE_MID(TRACE_INFO, ("Lost the race to create log thread"));
		TRACE_MID(TRACE_INFO, ("About to unlock logThreadSyncMutex"));
		threadMutexUnlock(logThreadSyncMutex);
		TRACE_RETURN(RSN_005, TT_SUCCESS);
	}
		
	/* create the log thread */
	TRACE_MID(TRACE_INFO, ("About to create log thread"));
	if (threadCreateThread((threadFunction_t)truncateLog,
						  (threadParameter_t)contextP,
						   &logThreadHandle) != THREAD_OK)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock logThreadSyncMutex"));
		threadMutexUnlock(logThreadSyncMutex);
		TRACE_RETURN(RSN_006, TT_FAILED);
	}
	TRACE_MID(TRACE_INFO, ("Log thread (%d) successfully created",
						   threadGetThreadId(logThreadHandle)));

	/* wait up to 10 seconds for child thread to signal us that he's up and running */
	TRACE_MID(TRACE_INFO, ("About to wait for log thread to initialise"));
	
	rc = threadCondTimedWait(logThreadSyncCondVar, 10000);

	if (rc != THREAD_OK && rc != THREAD_TIMED_OUT)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock logThreadSyncMutex"));
		threadMutexUnlock(logThreadSyncMutex);
		TRACE_RETURN(RSN_007, TT_FAILED);
	}

	/* check response from log thread */
	if (!logThreadInitialised)
	{
		sprintf(errorMessage, "Log thread failed to initialise.  Reason='%s'",
							  logThreadErrorMessage);
		TRACE_MID(TRACE_ERROR, ("%s", errorMessage));

		/* Output a log message */
		cciLog(&rc,                       /* returnCode */
			   CCI_LOG_ERROR,             /* type */
			   __FILE__,                  /* file */
			   __LINE__,                  /* line */
			   functionName,              /* function */
			   PLUGIN_MESSAGE_CATALOGUE,    /* messageSource */
			   MJB_LOG_THREAD_INIT_FAILED, /* messageNumber */ 
			   errorMessage,              /* traceText */
#if defined(_AIX)
			   brokerName,
#endif
			   errorMessage,
			   (char *)0);
	}
	else
	{
		logThreadCreated = TRUE;

		TRACE_MID(TRACE_INFO, ("Log thread initialised ok"));
	}

	TRACE_MID(TRACE_INFO, ("About to unlock logThreadSyncMutex"));
	threadMutexUnlock(logThreadSyncMutex);
	TRACE_RETURN(RSN_008, TT_SUCCESS);
}

/******************************************************************************/
/* deleteNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Nothing.                                                                  */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Called whenever an instance of a node object is destroyed.  Any resources  */
/* aquired by the node e.g. context or other resources, should be released    */
/* at this time.                                                              */
/*                                                                            */
/******************************************************************************/
void deleteNodeContext(CciContext *dataP)
{
	context_t *contextP = (context_t *)dataP;
	int rc;
    static char *functionName="deleteNodeContext()";
	char infoMessage[1000];
	char *nodeNameP,
		 *dotCharP,
	     mbLabelString[MAX_STRING_SIZE + 1];
	deployedAggregate_t *deployedAggregateP;

    TRACE_ENTRY("deleteNodeContext", ("dataP=%.8x", dataP));

	/* reset trace indentation just in case we've jumped here due to exception */
    TraceSetIndentationValue(2);  

	/* lock mutex just to be sure there's no race */
	TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
	threadMutexLock(aggregateChainMutex);

	/* free up the deployed aggregate that this node belongs to */
	deployedAggregateP = (deployedAggregate_t *)(contextP->nodeDataP);
    TRACE_MID(TRACE_INFO, ("Deleting deployed aggregate at %.8x",
				           deployedAggregateP));
    deleteAggregateInstances(contextP, deployedAggregateP);

	/* unlock aggregate chain mutex and allow alarm thread to run again */
	TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
	threadMutexUnlock(aggregateChainMutex);

	/* remove all entries on the timed out chain */

	/* timedOutAnchor */

	/* clear the attributes associated with this context */
    deleteAttributeTableEntries(contextP);

	/* clear the terminals associated with this context */
    deleteAllTerminalListEntries(contextP);

	/* 
	 * Finally, free up the context structure, but first create multibyte strings for
	 * inclusion in log message.
	 */
    UCToStr(contextP->nodeLabel,
		    sizeof(mbLabelString),
		    mbLabelString,
		    NULL);

	if (!wcscmp(contextP->nodeName, AGGREGATE_CREATE_NODE))
	{
		nodeNameP = AGGREGATE_CREATE_NODE_EXTERNAL_NAME_MB;
	}
	else
	{
		nodeNameP = AGGREGATE_REPLY_NODE_EXTERNAL_NAME_MB;
		/* 
		 * Modify the label so that it hides the existence of the SpARI node, e.g.
		 *     'SpAggregateReply2.ARI1' is modified to 'SpAggregateReply2'
		 */
		dotCharP = strchr(mbLabelString, '.');
		if (dotCharP)
		{
			*dotCharP = '\0';
		}
	}

	free(contextP->nodeName);
	free(contextP->nodeLabel);
    free(contextP);

	sprintf(infoMessage, "Successfully deleted node context for %s (%s)", 
		                  nodeNameP, mbLabelString);
	TRACE_MID(TRACE_INFO, ("%s", infoMessage));

    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   L"MQSeriesIntegratorPlugins", /* messageSource */
           MJB_NODE_CONTEXT_DELETED,  /* messageNumber */ 
		   infoMessage, /* traceText */
#if defined(_AIX)
		   brokerName,
#endif
           nodeNameP,
		   mbLabelString,
		   (char *)0);

	TRACE_RETURN_VOID(RSN_009);
}

/******************************************************************************/
/* deleteAggregateInstances                                                   */
/******************************************************************************/
void deleteAggregateInstances(IN context_t *contextP,
							  IN deployedAggregate_t *deployedAggregateP)
{
	aggregateInstance_t *aggregateP,
		                  *nextAggregateP;
	aggregateMember_t *memberP,
		              *nextMemberP;
	int rc;

	/*
	 * On entry, we assume caller has locked the aggregate chain mutex
	 */

    TRACE_ENTRY("deleteAggregateInstances", ("contextP=%.8x, deployedAggregateP=%.8x", 
		                                      contextP, deployedAggregateP));

	TRACE_MID(TRACE_INFO, ("processing deployed aggregate at %.8x",
						   deployedAggregateP));

	/* unchain the aggregate instances chain from the deployed aggregate block */
	aggregateP = deployedAggregateP->aggregateAnchor.nextP;
	deployedAggregateP->aggregateAnchor.nextP = NULL;

	while(aggregateP)
	{
		TRACE_MID(TRACE_INFO, ("processing aggregate descriptor at %.8x",
							   aggregateP));

		nextAggregateP = aggregateP->nextP;

		/* free up all aggregate members */
		memberP = aggregateP->memberAnchor.nextP;
		aggregateP->memberAnchor.nextP = NULL;

		while(memberP)
		{
			nextMemberP = memberP->nextP;
			TRACE_MID(TRACE_INFO, ("processing aggregate member at %.8x",
								   memberP));

			if (memberP->messageP)
			{
				TRACE_MID(TRACE_INFO, ("Releasing message storage at %.8x",
						                memberP->messageP));
				free(memberP->messageP);
	            TRACE_MID(TRACE_INFO, ("STORAGE(22): free  %.8x", memberP->messageP));
			}

			if (memberP->mqmdP)
			{
				TRACE_MID(TRACE_INFO, ("Releasing MQMD storage at %.8x",
						                memberP->mqmdP));
				free(memberP->mqmdP);
	            TRACE_MID(TRACE_INFO, ("STORAGE(23): free  %.8x", memberP->mqmdP));
			}

			TRACE_MID(TRACE_INFO, ("Removing index entry at %.8x",
									memberP->indexP));
			removeMsgIdIndex(memberP->indexP);

			TRACE_MID(TRACE_INFO, ("Releasing member storage at %.8x",
								   memberP));
			free(memberP);
            TRACE_MID(TRACE_INFO, ("STORAGE(24): free  %.8x", memberP));

			memberP = nextMemberP;
		}

		TRACE_MID(TRACE_INFO, ("Releasing aggregate descriptor storage at %.8x",
							   aggregateP));
		free(aggregateP);
        TRACE_MID(TRACE_INFO, ("STORAGE(25): free  %.8x", aggregateP));

		aggregateP = nextAggregateP;
	}

	/* decrement the number of nodes belonging to deployed aggregate */
	if (contextP->nodeType == AGGREGATE_CREATE)
	{
		deployedAggregateP->aggregateCreateCount--;
	}
	else
	{
		deployedAggregateP->aggregateReplyCount--;
	}

	/* if this is not the last node in the deployed aggregate, then do no more */
	if (deployedAggregateP->aggregateCreateCount ||
	    deployedAggregateP->aggregateReplyCount)
	{
		TRACE_MID(TRACE_INFO, ("Nodes still exist in deployed aggregate (create=%d, "
			                   "delete=%d).  Deployed aggregate will not be deleted",
							   deployedAggregateP->aggregateCreateCount,
							   deployedAggregateP->aggregateReplyCount));
		TRACE_RETURN_VOID(RSN_010);
	}

	/* unchain and free up the deployed aggregate structure */
	REMOVE_FROM_LINKED_LIST(deployedAggregateP);

	free(deployedAggregateP->name);
	if (deployedAggregateP->separatorString)
	{
		free(deployedAggregateP->separatorString);
	}

	TRACE_MID(TRACE_INFO, ("Releasing deployed aggregate storage at %.8x",
						   deployedAggregateP));
	free(deployedAggregateP);
	TRACE_MID(TRACE_INFO, ("STORAGE(26): free  %.8x", deployedAggregateP));

	/* Check if this was the last deployed aggregate we know of */
    if (deployedAggregateAnchor.nextP)
	{
		/* still more deployed aggregates, so just return */
		TRACE_RETURN_VOID(RSN_010);
	}

	/* There's no more deployed aggregates, so terminate alarm and log threads. */

	if (!alarmThreadShuttingDown && alarmThreadInitialised)
	{
		alarmThreadShuttingDown = TRUE;

		/* shutdown the alarm thread */
		shutdownRequested = TRUE;
		TRACE_MID(TRACE_INFO, ("Signalling alarm thread to shutdown"));
		threadCondSignal(alarmThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);

		/* wait for alarm to terminate before proceeding */
		rc = threadJoin(alarmThreadHandle);

		shutdownRequested = FALSE;

		TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
		threadMutexLock(aggregateChainMutex);
		if (rc == THREAD_OK)
		{
			alarmThreadCreated = FALSE;
			alarmThreadInitialised = FALSE;
			alarmThreadHandle = 0;
			alarmThreadShuttingDown = FALSE;
		}
	}

	if (logThreadInitialised && !logThreadShuttingDown)
	{
		logThreadShuttingDown = TRUE;

		/* shutdown the log thread */
		shutdownRequested = TRUE;
		TRACE_MID(TRACE_INFO, ("Signalling log thread to shutdown"));
		TRACE_MID(TRACE_INFO, ("About to lock logThreadWorkMutex"));
		threadMutexLock(logThreadWorkMutex);
		threadCondSignal(logThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock logThreadWorkMutex"));
		threadMutexUnlock(logThreadWorkMutex);

		/* wait for alarm to terminate before proceeding */
		rc = threadJoin(logThreadHandle);

		shutdownRequested = FALSE;

		if (rc == THREAD_OK)
		{
			logThreadCreated = FALSE;
			logThreadInitialised = FALSE;
			logThreadHandle = 0;
			logThreadShuttingDown = FALSE;
		}
	}

	TRACE_RETURN_VOID(RSN_011);
}

/******************************************************************************/
/* setAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context        (input)  - pointer to the current context                  */
/*  attributeName  (input)  - attribute name for which value is to be set     */
/*  attributeValue (input) -  value to be set for the attribute               */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Sets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a configuration request is received that attempts to    */
/* set the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/* This function has 3 responsibilities:                                      */
/*                                                                            */
/* 1. Verify value of attribute is correctly specified, otherwise a           */
/*    configuration exception should be thrown.                               */
/* 2. Store the attribute value in the context.                               */
/* 3. Throw a configuration exception if errors occur.                        */
/*                                                                            */
/******************************************************************************/
int setAttribute(CciContext *dataP,
				 CciChar *attributeName,
				 CciChar *attributeValue)
{
	int rc;
	context_t *contextP = (context_t *)dataP;

    TRACE_ENTRY("setAttribute", ("dataP=%.8x, attributeName=%S, "
		                         "attributeValue='%S'",
								 dataP, attributeName, attributeValue));

	/* call common plugin function */
	rc = commonSetAttribute(dataP,
		                    attributeName,
						    attributeValue);

	if (rc != CCI_SUCCESS)
	{
        TRACE_RETURN_TYPE(RSN_011, int, rc);
	}

	/* special processing for aggregate attributes */
    rc = processAttribute(contextP,
		                  attributeName,
		                  attributeValue);

    TRACE_RETURN_TYPE(RSN_012, int, rc);
}

/******************************************************************************/
/* processAttribute                                                           */
/******************************************************************************/
int processAttribute(IN context_t *contextP,
				     IN CciChar *attributeName,
				     IN CciChar *attributeValue)
{
	CciChar localSeparatorString[MAX_STRING_SIZE + 1];

	deployedAggregate_t *deployedAggregateP = NULL;
	int timeoutSeconds,
		sepStringLength,
		replyToQueueLength,
		replyDestLength,
		retCode;
	CciChar replyToQueue[MQ_Q_NAME_LENGTH + 1];
    CciChar replyDest[20 + 1];
	char msgBuffer[1000],
		 mbString[MAX_STRING_SIZE + 1],
		 *errorString;
	static char *functionName = "processAttribute";
	bool_t persistent,
		   instancesRecovered;
	messageFlow_t *mfP;

    TRACE_ENTRY("processAttribute", ("contextP=%.8x, attributeName=%S, "
		                             "attributeValue='%S'",
								     contextP, attributeName, attributeValue));

    /* 
	 * If this is the AggregateName attribute, then locate or create the
	 * aggregate descriptor and chain off the context.
	 */
	if (!wcscmp(attributeName, ATTR_AGGREGATE_NAME))
	{
	    TRACE_MID(TRACE_INFO, ("aggregate name is '%S'",
			                   attributeValue));

		/* 
		 * During message flow initialisation, it's unlikely that this code
		 * will be run concurrently.  However, just to be sure, we'll 
		 * serialise the code through this section.
		 */
		TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
		threadMutexLock(aggregateChainMutex);

		/* Find the current message flow entry */
		mfP = findMessageFlow(contextP->messageFlowName);

		if (!mfP)
		{
			/* The message flow entry should exist at this stage... */

			/* raise exception */
		}

		/* locate aggregate descriptor if it exists */
		deployedAggregateP = findDeployedAggregate(attributeValue);

		/* add new aggregate descriptor if one doesn't exist */
		if (!deployedAggregateP)
		{
			deployedAggregateP = 
				     (deployedAggregate_t *)malloc(sizeof(deployedAggregate_t));
			memset(deployedAggregateP, 0, sizeof(deployedAggregate_t));
		    TRACE_MID(TRACE_INFO, ("New deployed aggregate allocated at %.8x",
				                   deployedAggregateP));
	        TRACE_MID(TRACE_INFO, ("STORAGE(9): alloc %.8x", deployedAggregateP));


			/* copy in the name */
			deployedAggregateP->name = UCDup(attributeValue);

			TRACE_MID(TRACE_INFO, ("Linking to messageFlow entry at %.8x", mfP));
			deployedAggregateP->messageFlowEntryP = mfP;

			/* link deployed aggregate and current context */
			deployedAggregateP->contextP = contextP;
			contextP->nodeDataP = deployedAggregateP;

			/* add to the deployed aggregate chain */
			INSERT_AFTER_IN_LINKED_LIST(deployedAggregateAnchorP,
					                        deployedAggregateP);
		}
		else
		{
			/* ensure that this node is deployed in same message flow as remaining node(s) */
			if (deployedAggregateP->messageFlowEntryP != mfP)
			{
				/* raise an exception */
				TRACE_MID(TRACE_ERROR, ("Aggregate exists in more than one message flow"));
			}
		}

		if (contextP->nodeType == AGGREGATE_REPLY)
		{
			/* if this is not the first reply node, then error */
			if (deployedAggregateP->aggregateReplyCount)
			{
				sprintf(msgBuffer, "Aggregate %S is missing either an "
								   "AggregateCreate node or an AggregateReply "
								   "node", deployedAggregateP->name);
				TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

				/* throw an exception */
				cciThrowException(0,
								  CCI_FATAL_EXCEPTION,
								  __FILE__,
								  __LINE__,
								  functionName,
								  PLUGIN_MESSAGE_CATALOGUE,
								  MJB_TOO_MANY_REPLY_NODES, /* messageNumber */ 
								  msgBuffer,           /* traceText */
#if defined(_AIX)
		                          brokerName,
#endif
				                  /* convert aggregate name to multi byte string */
                                  UCToStr(deployedAggregateP->name,
						                  sizeof(mbString),
						                  mbString,
						                  NULL),
								  (char *)0);
			}

			/* bump the number of aggregate reply nodes */
			deployedAggregateP->aggregateReplyCount++;
			TRACE_MID(TRACE_INFO, ("AggregateReply node count for deployed "
									"aggregate '%S' set to %d",
									deployedAggregateP->name,
									deployedAggregateP->aggregateReplyCount));


			/* now retrieve any settings we have already processed */
			if (getIntegerAttribute(contextP, 
					                ATTR_REPLY_TIMEOUT_SECONDS,
					                &timeoutSeconds) == TT_SUCCESS)
			{
				deployedAggregateP->timeoutInterval = timeoutSeconds;
		        TRACE_MID(TRACE_INFO, ("%S value set", ATTR_REPLY_TIMEOUT_SECONDS));
			}

			if (getBooleanAttribute(contextP, 
					                ATTR_PERSISTENT,
					                &persistent) == TT_SUCCESS)
			{
				deployedAggregateP->persistent = persistent;
		        TRACE_MID(TRACE_INFO, ("%S value set to %s", ATTR_PERSISTENT,
					      deployedAggregateP->persistent ? "TRUE" : "FALSE"));
			}

			replyDestLength = sizeof(replyDest);
			if (getStringAttribute(contextP, 
					               ATTR_AGGREGATE_REPLY_DESTINATION,
								   &replyDestLength,
					               replyDest) == TT_SUCCESS)
			{
				deployedAggregateP->outputToDestinationList = FALSE;
				deployedAggregateP->outputToMQReply = FALSE;
				/* 
				 * Possible values are 
				 * - default
				 * - ReplyToQueue
				 * - DestinationList
				 */
				if (!wcscmp(replyDest, L"ReplyToQueue"))
				{
					deployedAggregateP->outputToMQReply = TRUE;
			        TRACE_MID(TRACE_INFO, ("outputToMQReply value is TRUE"));
				}
				else
				if (!wcscmp(replyDest, L"DestinationList"))
				{
					deployedAggregateP->outputToDestinationList = TRUE;
			        TRACE_MID(TRACE_INFO, ("outputToDestinationList value is TRUE"));
				}
				else
				{
			        TRACE_MID(TRACE_INFO, ("output value is default"));
				}
			}

			replyToQueueLength = sizeof(replyToQueue);
			if (getStringAttribute(contextP, 
					               ATTR_REPLY_TO_QUEUE,
								   &replyToQueueLength,
					               replyToQueue) == TT_SUCCESS)
			{
				wcscpy(deployedAggregateP->replyToQueueName,
					   replyToQueue);
		        TRACE_MID(TRACE_INFO, ("%S value set", ATTR_REPLY_TO_QUEUE));
			}

			sepStringLength = sizeof(localSeparatorString);
			if (getStringAttribute(contextP,
                                   ATTR_SEPARATOR_STRING,
						     	   &sepStringLength,
						           localSeparatorString) == TT_SUCCESS)
			{
				if (sepStringLength)
				{
					/* convert to multi byte string */
                    UCToStr(localSeparatorString,
						    sizeof(mbString),
						    mbString,
							NULL);
		            TRACE_MID(TRACE_INFO, ("sepStringLength is %d", sepStringLength));
					deployedAggregateP->separatorString = strdup(mbString);
					deployedAggregateP->separatorStringLength = sepStringLength;
					deployedAggregateP->separatorStringSet = TRUE;
                    TRACE_MID(TRACE_INFO, ("separator string value set to '%s'",
							               deployedAggregateP->separatorString));
				    TRACE_MID(TRACE_INFO, ("STORAGE(11): alloc %.8x", 
							               deployedAggregateP->separatorString));
				}
			}

			if (!deployedAggregateP->separatorStringSet)
			{
				/* process hex version */
				if (getHexAttribute(contextP,
                                    ATTR_SEPARATOR_STRING,
							   		&sepStringLength,
							        localSeparatorString) == TT_SUCCESS)
				{
					if (sepStringLength)
					{
						deployedAggregateP->separatorString = 
							(char *)malloc(sepStringLength);
				        TRACE_MID(TRACE_INFO, ("STORAGE(10): alloc %.8x", 
							                   deployedAggregateP->separatorString));

						memcpy(deployedAggregateP->separatorString, 
							   localSeparatorString,
							   sepStringLength);
						deployedAggregateP->separatorStringLength = sepStringLength;
						deployedAggregateP->separatorStringSet = TRUE;
	                    TRACE_MID(TRACE_INFO, ("separator string value set to..."));
	                    TRACE_MIDHEX(TRACE_INFO, deployedAggregateP->separatorString,
								     deployedAggregateP->separatorStringLength);
					}
				}
			}
		}
		else
		{
			/* bump the number of aggregate create nodes */
			deployedAggregateP->aggregateCreateCount++;
			TRACE_MID(TRACE_INFO, ("AggregateCreate node count for deployed "
									"aggregate '%S' set to %d",
									deployedAggregateP->name,
									deployedAggregateP->aggregateCreateCount));
		}

		/* chain off the current context */
		contextP->nodeDataP = (void *)deployedAggregateP;

		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);
	}
	else
	{
		/* update deployed aggregate if it exists */
		if (contextP->nodeDataP && contextP->nodeType == AGGREGATE_REPLY)
		{
			deployedAggregateP = (deployedAggregate_t *)(contextP->nodeDataP);

            TRACE_MID(TRACE_INFO, ("Updating existing deployed aggregate at %.8x",
				                   deployedAggregateP));

	        if (!wcscmp(attributeName, ATTR_REPLY_TIMEOUT_SECONDS) &&
			    getIntegerAttribute(contextP, 
					                ATTR_REPLY_TIMEOUT_SECONDS,
					                &timeoutSeconds) == TT_SUCCESS)
			{
				deployedAggregateP->timeoutInterval = timeoutSeconds;
	            TRACE_MID(TRACE_INFO, ("%S value updated", ATTR_REPLY_TIMEOUT_SECONDS));
			}

			replyDestLength = sizeof(replyDest);
			if (getStringAttribute(contextP, 
					               ATTR_AGGREGATE_REPLY_DESTINATION,
								   &replyDestLength,
					               replyDest) == TT_SUCCESS)
			{
				deployedAggregateP->outputToDestinationList = FALSE;
				deployedAggregateP->outputToMQReply = FALSE;
				/* 
				 * Possible values are 
				 * - default
				 * - ReplyToQueue
				 * - DestinationList
				 */
				if (!wcscmp(replyDest, L"ReplyToQueue"))
				{
					deployedAggregateP->outputToMQReply = TRUE;
			        TRACE_MID(TRACE_INFO, ("outputToMQReply value is TRUE"));
				}
				else
				if (!wcscmp(replyDest, L"DestinationList"))
				{
					deployedAggregateP->outputToDestinationList = TRUE;
			        TRACE_MID(TRACE_INFO, ("outputToDestinationList value is TRUE"));
				}
				else
				{
			        TRACE_MID(TRACE_INFO, ("output value is default"));
				}
			}

	        if (!wcscmp(attributeName, ATTR_PERSISTENT) &&
			    getBooleanAttribute(contextP, 
					                ATTR_PERSISTENT,
					                &persistent) == TT_SUCCESS)
			{
				deployedAggregateP->persistent = persistent;
		        TRACE_MID(TRACE_INFO, ("%S value set to %s", ATTR_PERSISTENT,
					      deployedAggregateP->persistent ? "TRUE" : "FALSE"));
			}

			replyToQueueLength = sizeof(replyToQueue);
	        if (!wcscmp(attributeName, ATTR_REPLY_TO_QUEUE) &&
			    getStringAttribute(contextP, 
					               ATTR_REPLY_TO_QUEUE,
								   &replyToQueueLength,
					               replyToQueue) == TT_SUCCESS)
			{
				wcscpy(deployedAggregateP->replyToQueueName,
					   replyToQueue);
	            TRACE_MID(TRACE_INFO, ("%S value updated", ATTR_REPLY_TO_QUEUE));
			}

			if (!(wcscmp(attributeName, ATTR_SEPARATOR_STRING) &&
				  wcscmp(attributeName, ATTR_HEX_SEPARATOR_STRING)) &&
				!deployedAggregateP->separatorStringSet)
			{

				/* process the separator string */
				if (!wcscmp(attributeName, ATTR_SEPARATOR_STRING))
				{
                    TRACE_MID(TRACE_INFO, ("'%S' value being processed", 
						                   ATTR_SEPARATOR_STRING));
					sepStringLength = sizeof(localSeparatorString);
					if (getStringAttribute(contextP,
                                           ATTR_SEPARATOR_STRING,
							     		   &sepStringLength,
								           localSeparatorString) == TT_SUCCESS)
					{
						if (sepStringLength)
						{
							/* convert to multi byte string */
							UCToStr(localSeparatorString,
								    sizeof(mbString),
									mbString,
									NULL);
							deployedAggregateP->separatorString = strdup(mbString);
							deployedAggregateP->separatorStringLength = sepStringLength;
							deployedAggregateP->separatorStringSet = TRUE;
	                        TRACE_MID(TRACE_INFO, ("separator string value set to '%s'",
								                   deployedAggregateP->separatorString));
						}
					}
				}
				else
				{
                    TRACE_MID(TRACE_INFO, ("'%S' value being processed", 
						                   ATTR_HEX_SEPARATOR_STRING));
					/* process hex version */
					if (getHexAttribute(contextP,
                                        ATTR_SEPARATOR_STRING,
							   		    &sepStringLength,
								        localSeparatorString) == TT_SUCCESS)
					{
						if (sepStringLength)
						{
							deployedAggregateP->separatorString = 
								(char *)malloc(sepStringLength);
							TRACE_MID(TRACE_INFO, ("STORAGE(12): alloc %.8x", 
												   deployedAggregateP->separatorString));
							memcpy(deployedAggregateP->separatorString, 
								   localSeparatorString,
								   sepStringLength);
							deployedAggregateP->separatorStringLength = sepStringLength;
							deployedAggregateP->separatorStringSet = TRUE;
	                        TRACE_MID(TRACE_INFO, ("separator string value set to..."));
	                        TRACE_MIDHEX(TRACE_INFO, deployedAggregateP->separatorString,
								         deployedAggregateP->separatorStringLength);
						}
					}
				}
			}
		}
	}

	/* make sure we've got at least one create node and at most one reply node */
    TRACE_MID(TRACE_INFO, ("contextP=%.8x, contextP->nodeDataP=%.8x",
		                    contextP, contextP->nodeDataP));
	deployedAggregateP = (deployedAggregate_t *)(contextP->nodeDataP);
	if (deployedAggregateP)
	{
        TRACE_MID(TRACE_INFO, ("Checking for valid aggregator configuration"));

		/* If both create and reply nodes are now present, make a note of this */
		if (!deployedAggregateP->aggregateCreateAndReplyExist &&
			 deployedAggregateP->aggregateReplyCount &&
			 deployedAggregateP->aggregateCreateCount)
		{
			deployedAggregateP->aggregateCreateAndReplyExist = TRUE;

			/* do recovery work here if not yet done... */
			if (deployedAggregateP->persistent && !deployedAggregateP->recoveryComplete)
			{
		        TRACE_MID(TRACE_INFO, ("About to recover existing instances for "
					                   "aggregate %S, messageFlow %S",
									   deployedAggregateP->name,
									   deployedAggregateP->messageFlowEntryP->name));

				/* lock the aggregate chain while its being modified */
				TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
				threadMutexLock(aggregateChainMutex);
				retCode = logRecoverInstances(contextP,
					                          deployedAggregateP,
											  &instancesRecovered,
											  &errorString);

				if (retCode != TT_SUCCESS)
				{
					TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
					threadMutexUnlock(aggregateChainMutex);
					sprintf(msgBuffer, "Failed to recover aggregate %S, error text is '%s'",
                                        deployedAggregateP->name,
										errorString);
					TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

					/* throw an exception */
					cciThrowException(0,
									  CCI_FATAL_EXCEPTION,
									  __FILE__,
									  __LINE__,
									  functionName,
									  PLUGIN_MESSAGE_CATALOGUE,
									  MJB_AGGREGATE_RECOVERY_FAILURE, /* messageNumber */ 
									  msgBuffer,           /* traceText */
#if defined(_AIX)
									  brokerName,
#endif
									  /* convert aggregate name to multi byte string */
									  UCToStr(deployedAggregateP->name,
											  sizeof(mbString),
											  mbString,
											  NULL),
									  errorString,
									  (char *)0);
				}

			    deployedAggregateP->recoveryComplete = TRUE;

				if (instancesRecovered)
				{
				/* 
				 * Give the alarm thread a chance to 'time out' any recovered
				 * aggregate instances for this deployed aggregate.
				 */
					TRACE_MID(TRACE_INFO, ("Signalling alarm thread to process recovered "
										   "aggregate instance(s)"));
					threadCondSignal(alarmThreadWorkCondVar);
				}

				TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
				threadMutexUnlock(aggregateChainMutex);
			}
		}

		if (deployedAggregateP->aggregateCreateAndReplyExist)
		{
			TRACE_MID(TRACE_INFO, ("Deployed aggregate '%S' contains a valid "
								   "configuration",
								   deployedAggregateP->name));
		}

		/* 
		 * If this is the first persistent deployed aggregate node encountered, create
		 * the log truncation thread
		 */
		if (deployedAggregateP->persistent && !logThreadCreated)
		{
			TRACE_MID(TRACE_INFO, ("Creating log thread"));
			retCode = createLogThread(contextP);

			if (retCode != TT_SUCCESS)
			{
				/* Output a log message */
				cciLog(0,                         /* returnCode */
					   CCI_LOG_ERROR,             /* type */
					   __FILE__,                  /* file */
					   __LINE__,                  /* line */
					   functionName,              /* function */
					   PLUGIN_MESSAGE_CATALOGUE,    /* messageSource */
					   MJB_LOG_THREAD_INIT_FAILED_NO_INSERT, /* messageNumber */ 
					   "Failed to create log truncation thread", /* traceText */
		#if defined(_AIX)
					   brokerName,
		#endif
					   (char *)0);
			
			
			}
		}
	}
								  
    TRACE_RETURN(RSN_013, TT_SUCCESS);
}

void showGlobals(IN deployedAggregate_t *deployedAggregateP)
{
    TRACE_ENTRY("showGlobals", ("deployedAggregateP=%.8x",deployedAggregateP));

	TRACE_MID(TRACE_INFO, ("timeoutInterval=%d", deployedAggregateP->timeoutInterval));
	TRACE_MID(TRACE_INFO, ("aggregateReplyCount=%d", deployedAggregateP->aggregateReplyCount));
	TRACE_MID(TRACE_INFO, ("aggregateCreateCount=%d", deployedAggregateP->aggregateCreateCount));
	TRACE_MID(TRACE_INFO, ("outputToMQReply=%s", deployedAggregateP->outputToMQReply ? "TRUE":"FALSE"));
	TRACE_MID(TRACE_INFO, ("outputToDestinationList=%s", deployedAggregateP->outputToDestinationList ? "TRUE":"FALSE"));
	TRACE_MID(TRACE_INFO, ("separatorStringSet=%s", deployedAggregateP->separatorStringSet ? "TRUE":"FALSE"));
	TRACE_MID(TRACE_INFO, ("persistent=%s", deployedAggregateP->persistent ? "TRUE":"FALSE"));
	TRACE_MID(TRACE_INFO, ("aggregateCreateAndReplyExist=%s", deployedAggregateP->aggregateCreateAndReplyExist ? "TRUE":"FALSE"));
	TRACE_MID(TRACE_INFO, ("replyToQueueName=%S", deployedAggregateP->replyToQueueName));
    TRACE_RETURN_VOID(RSN_014);
}

/******************************************************************************/
/* findDeployedAggregate                                                      */
/******************************************************************************/
deployedAggregate_t *findDeployedAggregate(IN CciChar *name)
{
	deployedAggregate_t *deployedAggregateP;
    TRACE_ENTRY("findDeployedAggregate", ("name='%S'", name));

	/* this function assumes that the aggregate chain lock is locked */
    deployedAggregateP = deployedAggregateAnchor.nextP;

	while(deployedAggregateP)
	{
		TRACE_MID(TRACE_INFO, ("Comparing with aggregate '%S'",
			                   deployedAggregateP->name));
		if (!wcscmp(deployedAggregateP->name, name))
		{
		    TRACE_MID(TRACE_INFO, ("Found a match at %.8x", deployedAggregateP));
			break;
		}
		deployedAggregateP = deployedAggregateP->nextP;
	}

	if (deployedAggregateP)
	{
	    TRACE_MID(TRACE_INFO, ("Returning deployed aggregate at %.8x",
			                   deployedAggregateP));
	}
	else
	{
	    TRACE_MID(TRACE_INFO, ("No matching deployed aggregate found"));
	}

    TRACE_RETURN_TYPE(RSN_015, deployedAggregate_t *, deployedAggregateP);
}

/******************************************************************************/
/* evaluate                                                                   */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Performs node processing. This function contains the main processing logic */
/* of the message flow node. It is invoked by the broker when a message is    */
/* received on one of the input terminals of an instance of a node object.    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Process the message in accordance with the values of any attribute on   */
/*    the node instance.                                                      */
/* 2. Process the message based on content if required.                       */
/* 3. Propagate the message to the appropriate output terminal                */
/* 4. Throw an exception if errors occur.                                     */
/*                                                                            */
/******************************************************************************/
void evaluate(CciContext *dataP,
		      CciMessage *destinationList,
		      CciMessage *exceptionList,
		      CciMessage *message)
{
	context_t *contextP;
	int rc;
	static char *functionName = "evaluate";
	char msgBuffer[200],
		 nodeName[100];

    TRACE_ENTRY("evaluate", ("dataP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 dataP, destinationList,
							 exceptionList, message));

	contextP = (context_t *)dataP;

	/* determine which function to call based on node name */
    TRACE_MID(TRACE_INFO, ("nodeName='%S'", contextP->nodeName));
	if (!wcscmp(contextP->nodeName, AGGREGATE_CREATE_NODE))
	{
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>> AggregateCreate >>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
		rc = evaluateAggregateCreate(contextP,
			                         destinationList,
								     exceptionList,
								     message);
	}
	else
	if (!wcscmp(contextP->nodeName, AGGREGATE_REPLY_NODE))
	{
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<< AggregateReply <<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
		rc = evaluateAggregateReply(contextP,
			                        destinationList,
							        exceptionList,
							        message);
	}
	else
	{
        TRACE_MID(TRACE_ERROR, ("Unrecognised node"));
		/* throw an exception here...*/
		sprintf(msgBuffer, "Unrecognised node ('%S') passed to evaluate()",
			               contextP->nodeName);
		sprintf(nodeName, "'%S'", contextP->nodeName);
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      PLUGIN_MESSAGE_CATALOGUE,
			              MJB_UNRECOGNISED_NODE,  /* messageNumber */ 
			              msgBuffer,              /* traceText */
#if defined(_AIX)
						  brokerName,
#endif
						  nodeName,
			              (char *)0);
	}

    TRACE_RETURN_VOID(RSN_016);
}

/******************************************************************************/
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Address of factory object.                                                */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine                                        */
/*                                                                            */
/* This function has 5 responsibilities:                                      */
/*                                                                            */
/* 1. create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/* 2. define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/* 3. define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/* 4. define input and output terminals supported by each node.               */
/* 5. return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */    
/*                                                                            */
/******************************************************************************/
#if defined(WIN32)
    #include <process.h>
#elif defined(_AIX)
    #include <unistd.h>
#elif defined(_SOLARIS)
    #include <unistd.h>
#endif
#if defined(WIN32)
void __declspec(dllexport) *bipGetMessageflowNodeFactory()
#elif defined(_AIX)
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
#endif
{
    static char *functionName="bipGetMessageflowNodeFactory()";
    void* factoryObject = NULL;
    CciChar factoryName[] = L"AggregatorPlugInFactory";
    CCI_EXCEPTION_ST exception_st;
    int rc;
    static CNI_VFT vftable;
	bool_t resourceCreated = TRUE;
	char *resourceName,
		 msgBuffer[200];

    TRACE_ENTRY("bipGetMessageflowNodeFactory", ("void"));
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));
    TRACE_MID(TRACE_INFO, ("<**** Start of aggregator deployment (pid=%d) ****>",
		                   getpid()));
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));

	/* create the global syncronisation resources we need here */

	/* create required mutexes */
	if (resourceCreated)
	{
        TRACE_MID(TRACE_INFO, ("About to create aggregateChainMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &aggregateChainMutex) != THREAD_OK)
		{
			resourceName = "aggregateChainMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadSyncMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &alarmThreadSyncMutex) != THREAD_OK)
		{
			resourceName = "alarmThreadSyncMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create logThreadSyncMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &logThreadSyncMutex) != THREAD_OK)
		{
			resourceName = "logThreadSyncMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create logThreadWorkMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &logThreadWorkMutex) != THREAD_OK)
		{
			resourceName = "logThreadWorkMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create uuidGenMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &uuidGenMutex) != THREAD_OK)
		{
			resourceName = "uuidGenMutex";
			resourceCreated = FALSE;
		}
	}

	/* create required condition variables */
	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadWorkCondVar"));
		if (threadCreateCondVar(aggregateChainMutex,
								&alarmThreadWorkCondVar) != THREAD_OK)
		{
			resourceName = "alarmThreadWorkCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadSyncCondVar"));
		if (threadCreateCondVar(alarmThreadSyncMutex,
								&alarmThreadSyncCondVar) != THREAD_OK)
		{
			resourceName = "alarmThreadSyncCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create logThreadSyncCondVar"));
		if (threadCreateCondVar(logThreadSyncMutex,
								&logThreadSyncCondVar) != THREAD_OK)
		{
			resourceName = "logThreadSyncCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create logThreadWorkCondVar"));
		if (threadCreateCondVar(logThreadWorkMutex,
								&logThreadWorkCondVar) != THREAD_OK)
		{
			resourceName = "logThreadWorkCondVar";
			resourceCreated = FALSE;
		}
	}

	/* process any errors that have occurred */
	if (!resourceCreated)
	{
		sprintf(msgBuffer, "Aggregate Plugin Factory failed to create resource %s",
						   resourceName);
		/* generate an exception here... */
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      PLUGIN_MESSAGE_CATALOGUE,
			              MJB_RESOURCE_CREATION_FAILED,  /* messageNumber */ 
			              msgBuffer, /* traceText */
#if defined(_AIX)
						  brokerName,
#endif
			              "AggregatorPlugInFactory",
						  resourceName,
			              (char *)0);
	}


    /* Setup virtual function table with addresses of implementation functions */
    memset(&vftable, 0, sizeof(vftable));
    vftable.iFpCreateNodeContext     = createNodeContext;
    vftable.iFpDeleteNodeContext     = deleteNodeContext;
    vftable.iFpGetAttributeName      = getAttributeName;
    vftable.iFpSetAttribute          = setAttribute;
    vftable.iFpGetAttribute          = getAttribute;
    vftable.iFpEvaluate              = evaluate;


    /* Create the Node Factory for this plug-in */
    factoryObject = cniCreateNodeFactory(&rc, factoryName);
    if (factoryObject == CCI_NULL_ADDR)
	{
        if (rc == CCI_EXCEPTION)
		{
            cciGetLastExceptionData(&rc, 
				                    &exception_st);
            /* Any local error handling may go here */
            cciRethrowLastException(&rc);
		}
        /* Any further local error handling can go here */
        return(factoryObject);
	}
    /* Define the AggregateCreate and AggregateReply nodes supported by our factory */

	/*
	 * Define the AggregateCreate node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject,
					   AGGREGATE_CREATE_NODE,
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create AggregateCreate node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "Aggregator Plugin Factory failed to create AggregateCreate node", /* traceText */
#if defined(_AIX)
						  brokerName,
#endif
			              "AggregatorPlugInFactory",
			              AGGREGATE_CREATE_NODE_EXTERNAL_NAME_MB,
			              (char *)0);
	}

	/*
	 * Define the AggregateReply node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject, 
					   AGGREGATE_REPLY_NODE, 
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create AggregateReply node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "Aggregator Plugin Factory failed to create AggregateReply node", /* traceText */
#if defined(_AIX)
						  brokerName,
#endif
			              "AggregatorPlugInFactory",
			              AGGREGATE_REPLY_NODE_EXTERNAL_NAME_MB,
			              (char *)0);
	}

	/* create a thread key for the logging function */
	if (!loggingKey) 
	{
		rc = threadGetSlotKey(&loggingKey);
        TRACE_MID(TRACE_INFO, ("Created thread slot for recovery queue details (%.8x)",
			                   loggingKey));
	}
	else
	{
        TRACE_MID(TRACE_INFO, ("Thread slot already created (%.8x)", loggingKey));
	}

    /* Log a success message */
    TRACE_MID(TRACE_INFO, ("About to issue cciLog success message"));
    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   PLUGIN_MESSAGE_CATALOGUE,  /* messageSource */
#if defined(_AIX)
           MJB_NODE_FACTORY_CREATED_AIX,  /* messageNumber */ 
#else
           MJB_NODE_FACTORY_CREATED,  /* messageNumber */ 
#endif
		   "Aggregator Plugin Factory loading", /* traceText */
		   "AggregatorPlugInFactory",
		   "AggregatorPlugInFactory",
		   (char *)0);

    /* Return address of this factory object to the broker */
    TRACE_RETURN_TYPE(RSN_017, void *, factoryObject);
}

