/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : garbage.c                                                */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Provides global storage garbage collection functions for */
/*                   use by postit node.                                      */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <postit.h> /* Specific header for this dll                            */

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <threads.h>
#include <time.h>

#define DEFAULT_SLEEP_SECONDS 20
/******************************************************************************/
/* sharedMemoryGarbageCollection                                              */
/******************************************************************************/
threadFunction_t sharedMemoryGarbageCollection(void *dataP)
{
	time_t currentTime;
	int timeToWait,
		numPostitsFound,
		rc,
		brokerPercentFree = 100,
		platformPercentFree = 100,
		minPercentFree;
	double fPercentFree;
	globalPostitCreateInstance_t *deadAnchorBrokerP = NULL,
		                         *deadAnchorPlatformP = NULL,
		                         *currentInstanceP,
								 *nextInstanceP;
    static char *functionName="sharedMemoryGarbageCollection";

    TRACE_ENTRY("sharedMemoryGarbageCollection", ("dataP=%.8x", dataP));

	/* let the parent thread know that we're up and running */
	shmemThreadInitialised = TRUE;

	/* lock the alarm thread sync mutex and post the parent thread */
    TRACE_MID(TRACE_INFO, ("About to signal parent for initialisation complete"));
	threadMutexLock(shmemThreadSyncMutex);
	threadCondSignal(shmemThreadSyncCondVar);
	threadMutexUnlock(shmemThreadSyncMutex);

	/* address the dead chains */
	if (!deadAnchorBrokerP && sharedMemoryCBP_Broker)
	{
		deadAnchorBrokerP = &sharedMemoryCBP_Broker->globalDeadInstanceAnchor;
	}
	if (!deadAnchorPlatformP && sharedMemoryCBP_Platform)
	{
		deadAnchorPlatformP = &sharedMemoryCBP_Platform->globalDeadInstanceAnchor;
	}

	/* 
	 * Periodically check the global storage dead chain for any postits to be removed
	 */
	threadMutexLock(shmemThreadWorkMutex);

	/* keep processing global storage dead chain until requested to shutdown */
	while(!shutdownShmemRequested)
	{
		/* 
		 * Address the dead chains - another memory segment might have been added since we
		 * last checked
		 */
		if (!deadAnchorBrokerP && sharedMemoryCBP_Broker)
		{
			deadAnchorBrokerP = &sharedMemoryCBP_Broker->globalDeadInstanceAnchor;
		}
		if (!deadAnchorPlatformP && sharedMemoryCBP_Platform)
		{
			deadAnchorPlatformP = &sharedMemoryCBP_Platform->globalDeadInstanceAnchor;
		}

		/* take a peek at the Broker dead chain (dirty read) to avoid locking shared memory */
		numPostitsFound = 0;
		if (deadAnchorBrokerP) {
			// get the current heap stats 
			brokerPercentFree = shmGetPercentFree(TRUE,  // PRETEND we've got the lock
			                                      TRUE); // scope is broker 
		if (deadAnchorBrokerP->nextGlobalP != SHM_NULL_HANDLE) {
			/* 
			 * Ok, looks like theres something on the chain, so lock global storage 
			 * and check again.
			 */
			/* 
			 * The postit chain and the shared memory dead chain are related in that if we find
			 * anything to release on the dead chain, we also need to remove it from the
			 * local chains, so we always take the postitChainMutex first to avoid deadlock.
			 */
			TRACE_MID(TRACE_INFO, ("Processing Broker dead chain"));
			TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
			threadMutexLock(postitChainMutex);

			TRACE_MID(TRACE_INFO, ("About to take Broker global lock"));
			shmLock(TRUE);  // take the Broker scope lock

			// get the actual current heap stats before we start freeing storage
			brokerPercentFree = shmGetPercentFree(TRUE,  // we've got the lock
			                                      TRUE); // scope is broker 

			currentInstanceP = SHM_HANDLE2ADDRESS(deadAnchorBrokerP->nextGlobalP, 
				                                  globalPostitCreateInstance_t *,
												  sharedMemorySegment_Broker);
			while(currentInstanceP)
			{
				nextInstanceP = SHM_HANDLE2ADDRESS(currentInstanceP->nextGlobalP,
					                               globalPostitCreateInstance_t *,
												   sharedMemorySegment_Broker);

                /* skip this postit if it doesn't belong to our execution group */
				if (currentInstanceP->executionGroupPID != currentPID)
				{
					TRACE_MID(TRACE_INFO, ("Postit owned by process %d - postit ignored", 
						                    currentInstanceP->executionGroupPID));
				    currentInstanceP = nextInstanceP;
					continue;
				}

				/* process one of 'our' postits */
				numPostitsFound++;
				TRACE_MID(TRACE_INFO, ("We own this postit, so unchain and free it"));

				// all postits on the broker dead chain will already have been removed from
				// global chains.  Just need to remove from the dead chain.
				TRACE_MID(TRACE_INFO, ("currentInstanceP=%.8x", currentInstanceP));
				{
					// turn handles into pointers
					shmemHandle_t handle;
					globalPostitCreateInstance_t *gP;
					handle = currentInstanceP->prevGlobalP;
					if (handle==SHM_NULL_HANDLE) {
				        TRACE_MID(TRACE_INFO, ("currentInstanceP=prevGlobalP is 00000000"));
					}
					else {
						gP = SHM_HANDLE2ADDRESS(currentInstanceP->prevGlobalP, globalPostitCreateInstance_t *, sharedMemorySegment_Broker);
				        TRACE_MID(TRACE_INFO, ("currentInstanceP->prevGlobalP is %.8x", gP));
					}
					handle = currentInstanceP->nextGlobalP;
					if (handle==SHM_NULL_HANDLE) {
				        TRACE_MID(TRACE_INFO, ("currentInstanceP->nextGlobalP is 00000000"));
					}
					else {
						gP = SHM_HANDLE2ADDRESS(currentInstanceP->nextGlobalP, globalPostitCreateInstance_t *, sharedMemorySegment_Broker);
				        TRACE_MID(TRACE_INFO, ("currentInstanceP->nextGlobalP is %.8x", gP));
					}
				}

				REMOVE_FROM_GLOBAL_POSTIT_CHAIN(currentInstanceP, sharedMemorySegment_Broker);

				/* remove from local chains too...set scope to local */
				currentInstanceP->scope = SCOPE_EXECUTION_GROUP;
                unchainPostit((postitCreateInstance_t *)currentInstanceP);

				/* finally, free up the shared memory */
				rc = shmFree(currentInstanceP, TRUE, /* we've already got the shmem lock */
					                           TRUE); /* broker scope storage */ 
		        CHECK_SHMFREE_ERROR(rc, 
			                        functionName, 
						            platformIsUnix);  

				currentInstanceP = nextInstanceP;
			}

			TRACE_MID(TRACE_INFO, ("About to release Broker global lock"));
			shmUnlock(TRUE);  // free the Broker scope lock

			TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
			threadMutexUnlock(postitChainMutex);
		}
		}

		/* take a peek at the Platform dead chain (dirty read) to avoid locking shared memory */
		numPostitsFound = 0;
		if (deadAnchorPlatformP) {
			// get the current heap stats 
			platformPercentFree = shmGetPercentFree(TRUE,  // PRETEND we've got the lock
			                                        FALSE); // scope is platform NOT broker 
		if (deadAnchorPlatformP->nextGlobalP != SHM_NULL_HANDLE)
		{
			/* 
			 * Ok, looks like theres something on the chain, so lock global storage 
			 * and check again.
			 */
			/* 
			 * The postit chain and the shared memory dead chains are related in that if we find
			 * anything to release on the dead chain, we also need to remove it from the
			 * local chains, so we always take the postitChainMutex first to avoid deadlock.
			 */
			TRACE_MID(TRACE_INFO, ("Processing Platform dead chain"));
			TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
			threadMutexLock(postitChainMutex);

			TRACE_MID(TRACE_INFO, ("About to take Platform global lock"));
			shmLock(FALSE);  // take the Platform scope lock

			// get the actual current broker heap stats before we start freeing storage
			platformPercentFree = shmGetPercentFree(TRUE,  // we've got the lock
			                                        FALSE); // scope is platform NOT broker 

			currentInstanceP = SHM_HANDLE2ADDRESS(deadAnchorPlatformP->nextGlobalP, 
				                                  globalPostitCreateInstance_t *,
												  sharedMemorySegment_Platform);
            while(currentInstanceP)
			{
				nextInstanceP = SHM_HANDLE2ADDRESS(currentInstanceP->nextGlobalP,
					                               globalPostitCreateInstance_t *,
												   sharedMemorySegment_Platform);

                /* skip this postit if it doesn't belong to our execution group */
				if (currentInstanceP->executionGroupPID != currentPID)
				{
					TRACE_MID(TRACE_INFO, ("Postit owned by process %d - postit ignored", 
						                    currentInstanceP->executionGroupPID));
				    currentInstanceP = nextInstanceP;
					continue;
				}

				/* process one of 'our' postits */
				numPostitsFound++;
				TRACE_MID(TRACE_INFO, ("We own this postit, so unchain and free it"));

				// all postits on the broker dead chain will already have been removed from
				// global chains.  Just need to remove from the dead chain.
				TRACE_MID(TRACE_INFO, ("currentInstanceP=%.8x", currentInstanceP));
				{
					// turn handles into pointers
					shmemHandle_t handle;
					globalPostitCreateInstance_t *gP;
					handle = currentInstanceP->prevGlobalP;
					if (handle==SHM_NULL_HANDLE) {
				        TRACE_MID(TRACE_INFO, ("currentInstanceP->prevGlobalP is 00000000"));
					}
					else {
						gP = SHM_HANDLE2ADDRESS(currentInstanceP->prevGlobalP, globalPostitCreateInstance_t *, sharedMemorySegment_Platform);
				        TRACE_MID(TRACE_INFO, ("currentInstanceP=prevGlobalP is %.8x", gP));
					}
					handle = currentInstanceP->nextGlobalP;
					if (handle==SHM_NULL_HANDLE) {
				        TRACE_MID(TRACE_INFO, ("currentInstanceP->nextGlobalP is 00000000"));
					}
					else {
						gP = SHM_HANDLE2ADDRESS(currentInstanceP->nextGlobalP, globalPostitCreateInstance_t *, sharedMemorySegment_Platform);
				        TRACE_MID(TRACE_INFO, ("currentInstanceP->nextGlobalP is %.8x", gP));
					}
				}

				REMOVE_FROM_GLOBAL_POSTIT_CHAIN(currentInstanceP, sharedMemorySegment_Platform);

				/* remove from local chains too...set scope to local */
				currentInstanceP->scope = SCOPE_EXECUTION_GROUP;
                unchainPostit((postitCreateInstance_t *)currentInstanceP);

				/* finally, free up the shared memory */
				rc = shmFree(currentInstanceP, TRUE, /* we've already got the shmem lock */
					                           FALSE); /* platform, not broker scope storage */ 
		        CHECK_SHMFREE_ERROR(rc, 
			                        functionName, 
						            platformIsUnix);  

				currentInstanceP = nextInstanceP;
			}

			TRACE_MID(TRACE_INFO, ("About to release Platform global lock"));
			shmUnlock(FALSE);  // free the Platform scope lock

			TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
			threadMutexUnlock(postitChainMutex);
		}
		}

		/* go to sleep for while... */
		/* reset the garbage thread to pop at next expiry interval */
		time(&currentTime);
        TRACE_MID(TRACE_INFO, ("currentTime=%d", currentTime));

		// calculate time to wait - this is directly linked to the amount of free storage left.
		// The more free storage remaining, the longer we sleep and vice versa.
		minPercentFree = platformPercentFree < brokerPercentFree ? platformPercentFree : brokerPercentFree; 

		fPercentFree = (double)(minPercentFree * DEFAULT_SLEEP_SECONDS) / (double)100;

        //TRACE_MID(TRACE_INFO, ("fPercentFree=%f", fPercentFree));

		timeToWait = (int)(fPercentFree);
		// round up the result
		if (fPercentFree - timeToWait >= 0.5) timeToWait++;

		if (timeToWait == 0) {
			timeToWait = 1;  // make sure we wait at least one second
		}
        TRACE_MID(TRACE_INFO, ("Time to wait set to %d seconds", timeToWait));

        TRACE_MID(TRACE_INFO, ("Next expiry interval time is %d", currentTime+timeToWait));

		// convert to milliseconds
		timeToWait *= 1000;

	    /* wait for next alarm to pop or a new element to appear on the chain */
        TRACE_MID(TRACE_INFO, ("Garbage collection thread about to sleep for expiry interval"));
		sharedMemoryGarbageCollectorHasRun = TRUE;
		rc = threadCondTimedWait(shmemThreadWorkCondVar, timeToWait);
		if (rc == THREAD_TIMED_OUT) {
            TRACE_MID(TRACE_INFO, ("Garbage collection thread woken normally"));
		}
		else {
            TRACE_MID(TRACE_INFO, ("Garbage collection thread woken prematurely"));
		}
	}

    TRACE_MID(TRACE_INFO, ("Shared memory garbage collection thread terminating"));

	threadMutexUnlock(shmemThreadWorkMutex);

    TRACE_RETURN_TYPE(RSN_001, FUNCTION_RETURN_TYPE, 0);
}

